"""
Author: Elias Vanderstuyft
Licence: GPL v3
TODOs:
    - Try to find/generate a FFE file that contains a CustomForce effect
    - Commented TODOs
    - Implement the Write part (shouldn't be too difficult, is inverse operation of Read part)
First release: Tuesday, February 18, 2014
Last edit: Tuesday, February 18, 2014
"""

import struct




##########################     Constants    ##########################

NULL = 0x00

GUIDs = {
    "\x20\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_ConstantForce",
    "\x21\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_RampForce",
    "\x22\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Square",
    "\x23\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Sine",
    "\x24\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Triangle",
    "\x25\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_SawtoothUp",
    "\x26\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_SawtoothDown",
    "\x27\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Spring",
    "\x28\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Damper",
    "\x29\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Inertia",
    "\x2A\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Friction",
    "\x2B\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_CustomForce"
        }




##########################     RIFF File Format Logic and Datatypes    ##########################

class DataTypes:
    
    dataTypes = {
            "LONG"      : (4, ('<', 'l')),
            "DWORD"     : (4, ('<', 'i')),
            "CHAR"      : (1, str)
                }
    
    @staticmethod
    def getTypeAndMultiplicity(dataTypeMult):
        if dataTypeMult in DataTypes.dataTypes:
            return dataTypeMult, -1
        elif dataTypeMult[-1] == '*' and dataTypeMult[:-1] in DataTypes.dataTypes:
            return dataTypeMult[:-1], 0
        else:
            lBracket = dataTypeMult.find("[")
            rBracket = dataTypeMult.find("]")
            if min(lBracket, rBracket) == -1:
                raise ValueError("Unknown multiplicity of datatype '%s'" % dataTypeMult)
            dataType = dataTypeMult[:lBracket]
            mult = eval(dataTypeMult[lBracket+1:rBracket])
            if dataType in DataTypes.dataTypes:
                return dataType, mult
            else:
                raise NameError("Unknown type ('%s') of datatype '%s'" % (dataType, dataTypeMult))
    
    @staticmethod
    def unpack(inp, offset, dataType):
        dataType, mult = DataTypes.getTypeAndMultiplicity(dataType)
        
        if mult == -1:
            realMult = 1
        elif mult == 0:
            realMult = float(len(inp) - offset) / DataTypes.dataTypes[dataType][0]
            if realMult % 1.:
                raise ValueError( "%s is not a whole multiple of %s (= sizeof('%s'))" % 
                                    (len(inp) - offset, DataTypes.dataTypes[dataType][0], dataType) )
            else:
                realMult = int(realMult)
        else:
            realMult = mult
        sizeToUnpack = realMult * DataTypes.dataTypes[dataType][0]
        if sizeToUnpack > len(inp) - offset:
            raise ValueError( "Cannot unpack %s instance(s) of '%s' (size %s), total size == %s != %s == size of data to unpack" % 
                                (realMult, dataType, DataTypes.dataTypes[dataType][0], sizeToUnpack, len(inp) - offset) )
        
        end = offset + sizeToUnpack
        text = inp[offset:end]
        if type(DataTypes.dataTypes[dataType][1]) == tuple:
            out = struct.unpack(DataTypes.dataTypes[dataType][1][0] + DataTypes.dataTypes[dataType][1][1] * realMult, text)
        else:
            def splitStrIntoEqualParts(s, n):
                return map( ''.join, zip(*([iter(s)] * n)) )
            out = map(DataTypes.dataTypes[dataType][1], splitStrIntoEqualParts(text, DataTypes.dataTypes[dataType][0]))
        if len(out) != realMult:
            raise ValueError( "Unpacking outputs %s instance(s) of '%s', instead of the expected amount of %s " % 
                                (len(out), dataType, realMult) )
        
        if dataType == "CHAR": # Special case for strings = char array
            return ''.join(out), end
        
        if mult == -1:
            return out[0], end
        else:
            return out, end
    
    @staticmethod
    def pack():
        print ("Not yet implemented.")


class RIFFChunk:
    
    def __init__(self, label = "", noLengthInHeader = False, length = 0, dontUnpackData = False, dataTypes = {}, dataTypeOrder = [], lastDataTypeIsArray = False):
        self.expectedLabel = label
        self.noLengthInHeader = noLengthInHeader
        self.expectedLength = length
        self.dontUnpackData = dontUnpackData
        if not dontUnpackData:
            self.setDataTypes(dataTypes, dataTypeOrder)
        self.lastDataTypeIsArray = lastDataTypeIsArray
        
        self.label = ""
        self.globalOffset = 0
        self.length = 0
        self.dataText = ""
        self.data = {}
        self.setAdditionalRepresentation("")
        self.parent = None
        self.excludedMembers = []
    
    def setDataTypes(self, dataTypes, dataTypeOrder):
        self.dataTypes = dataTypes
        self.dataTypeOrder = dataTypeOrder
        for member in self.dataTypeOrder:
            isChunk = (type(self.dataTypes[member]) != str)
            self.check( not isChunk or issubclass(self.dataTypes[member], RIFFChunk), 
                        "'%s' (associated with %s) is not a subclass of 'RIFFChunk'" % 
                            ((isChunk and self.dataTypes[member].__name__) or self.dataTypes[member], member) )
    
    def setParent(self, parent):
        self.parent = parent
    
    def alert(self, message, type = "Warning"):
        print ("'%s' at global offset %s: %s, %s" % (self.__class__.__name__, hex(self.globalOffset), type, message))
    
    def check(self, condition, message):
        try:
            assert condition
            return True
        except AssertionError:
            self.alert(message)
            return False
    
    def checkMember(self, member, expectedValue):
        return self.check( self.data[member] == expectedValue,
                            "data[\"%s\"] == %s != %s (expected value)" %
                                (member, self.data[member], repr(expectedValue)) )
    
    def checkMemberNot(self, member, forbiddenValue):
        return self.check( self.data[member] != forbiddenValue,
                            "data[\"%s\"] == %s (forbidden value)" %
                                (member, self.data[member]) )
    
    def checkMemberRange(self, member, start, end):
        return self.check( start <= self.data[member] <= end,
                            "data[\"%s\"] == %s is not in (expected) range [%s, %s]" %
                                (member, self.data[member], start, end) )
    
    def read(self, inp, offset, globalOffset = 0):
        self.globalOffset = globalOffset
        
        if self.expectedLabel:
            self.label, offset = DataTypes.unpack(inp, offset, "CHAR[4]")
            self.check( self.label == self.expectedLabel, 
                        "inp[%s:%s+4] == %s != %s == expected label" % 
                            (globalOffset+offset-4, globalOffset+offset-4, repr(self.label), repr(self.expectedLabel)) )
        
        if self.noLengthInHeader:
            self.length = self.expectedLength
        else:
            self.length, offset = DataTypes.unpack(inp, offset, "DWORD")
            if self.expectedLength:
                self.check( self.length == self.expectedLength, 
                            "at inp[%s:%s+4]: Chunk length == %s != %s" % 
                                (globalOffset+offset-4, globalOffset+offset-4, self.length, self.expectedLength) )
        
        self.globalOffset = globalOffset + offset
        self.dataText = inp[offset:offset + self.length]
        self.check( self.length == len(self.dataText), 
                    "data of chunk starting at offset %s: Real chunk length == %s != %s == Declared/Advertised chunk length" % 
                        (self.globalOffset, len(self.dataText), self.length) )
        
        if not self.dontUnpackData:
            self.unpackData()
        
        return offset + len(self.dataText)
    
    def unpackData(self):
        def unpack(member, offset):
            isChunk = (type(self.dataTypes[member]) != str)
            try:
                if isChunk:
                    data = self.dataTypes[member]()
                    data.setParent(self)
                    return data, data.read(self.dataText, offset, self.globalOffset)
                else:
                    return DataTypes.unpack(self.dataText, offset, self.dataTypes[member])
            except Exception as exception:
                import sys, traceback
                self.alert( "at unpacking %s '%s' at offset %s:" % 
                                (member, (isChunk and self.dataTypes[member].__name__) or self.dataTypes[member], hex(self.globalOffset+offset)), 
                            "Error" )
                traceback.print_exc(file = sys.stdout)
                raise exception
        
        offset = 0
        if self.lastDataTypeIsArray:
            dataTypeOrder = self.dataTypeOrder[:-1]
        else:
            dataTypeOrder = self.dataTypeOrder
        
        for member in dataTypeOrder:
            if member in self.excludedMembers:
                continue
            if self.check( offset < len(self.dataText),
                            "Ran out of data (offset == %s >= %s == len(dataText)) to unpack member '%s'" % 
                                (offset, len(self.dataText), member) ):
                self.data[member], offset = unpack(member, offset)
        
        if self.lastDataTypeIsArray and not self.dataTypeOrder[-1] in self.excludedMembers:
            member = self.dataTypeOrder[-1]
            self.data[member] = []
            while offset < len(self.dataText):
                preOffset = offset
                data, offset = unpack(member, offset)
                self.data[member].append(data)
                if offset == preOffset:
                    break
        
        self.check( offset == len(self.dataText),
                    "length of available data == %s != %s == expected length of all datatypes" % 
                        (len(self.dataText), offset) )
    
    def setAdditionalRepresentation(self, additionalRepr):
        self.additionalRepr = additionalRepr
    
    def reprUnpackedData(self, indent = 0):
        def reprUnpacked(dataMember, indent):
            if isinstance(dataMember, RIFFChunk):
                return dataMember.reprUnpackedData(indent)
            else:
                return "%s\n" % repr(dataMember)
        
        out = "%s%s (offset: %s)\n" % \
                ( self.__class__.__name__, (" (%s)"%self.additionalRepr)*(self.additionalRepr!=""), 
                    hex(self.globalOffset - 4 * ((self.expectedLabel!="") + (not self.noLengthInHeader))) )
        
        indent += 1
        for member in self.dataTypeOrder:
            out += "%s+-> %s = " % (indent * '\t', member)
            if member in self.data:
                if type(self.data[member]) in (list, tuple):
                    out += "\n"
                    indent += 1
                    out += "%s[\n" % (indent * '\t')
                    for i in self.data[member]:
                        out += "%s  %s" % (indent * '\t', reprUnpacked(i, indent))
                    out += "%s]\n" % (indent * '\t')
                    indent -= 1
                else:
                    out += reprUnpacked(self.data[member], indent)
            else:
                out += "<missing>\n"
        
        return out
    
    def write(self):
        print ("Not yet implemented.")
    
    def packData(self):
        print ("Not yet implemented.")





##########################     File Format Specification    ##########################

"""

Available types
    Generic
        DWORD
        LONG
        CHAR[length]
    Arrays
        <type>*
    Chunks
        - Should be a subclass of RIFFChunk


RIFFChunk summary

    FFEFile
        label
            "RIFF"
        noSize
        len
            len(inp)
        data
            "header1" : "CHAR[0x4]"
            "ffeHeader" : FFEHeader
            "fileEffect" : FileEffectChunk
        lastDataTypeIsArray

    FFEHeader
        label
            "FORC"
        noSize
        len
            0x10
        data
            "header2" : "CHAR[0x10]"

    FileEffectChunk
        label
            "LIST"
        data
            "effectStringChunk" : EffectStringChunk
            "lpDiEffect" : EffectChunk
            "GuidEffect" : "CHAR[0x10]"
            "unknownDWORD1" : "DWORD"
            "axesChunk" : AxesChunk
            "directionChunk" : DirectionChunk
            "envelopeChunk" : EnvelopeChunk
            "effectTypeSpecificChunk" : EffectTypeSpecificChunk
            "customForceDataChunk" : CustomForceDataChunk    (Warning: this is hypothetical, not tested because FEdit can't generate CustomForce effects)

    EffectStringChunk
        label
            "efct"
        noSize
        len
            0x40
        data
            "szFriendlyName" : "CHAR[0x40]"

    EffectChunk
        noSize
        len
            0x38
        data
            "dwSize" : "DWORD"
            "dwFlags" : "DWORD"
            "dwDuration" : "DWORD"
            "dwSamplePeriod" : "DWORD"
            "dwGain" : "DWORD"
            "dwTriggerButton" : "DWORD"
            "dwTriggerRepeatInterval" : "DWORD"
            "cAxes" : "DWORD"
            "rgdwAxes" : "LONG"
            "rglDirection" : "LONG"
            "lpEnvelope" : "LONG"
            "cbTypeSpecificParams" : "DWORD"
            "lpvTypeSpecificParams" : "LONG"
            "dwStartDelay" : "DWORD"

    AxesChunk
        label
            "data"
        data
            "rgdwAxes" : "DWORD*"

    DirectionChunk
        label
            "data"
        data
            "rglDirection" : "LONG*"

    EnvelopeChunk
        label
            "data"
        len
            0x14
        data
            "dwSize" : "DWORD"
            "dwAttackLevel" : "DWORD"
            "dwAttackTime" : "DWORD"
            "dwFadeLevel" : "DWORD"
            "dwFadeTime" : "DWORD"

    EffectTypeSpecificChunk
        label
            "data"
        dontUnpack
        len
            Case GUID_ConstantForce:
                0x4
            Case GUID_RampForce:
                0x8
            Case GUID_Square, GUID_Sine, GUID_Triangle, GUID_SawtoothUp, GUID_SawtoothDown:
                0x10
            Case GUID_Spring, GUID_Damper, GUID_Inertia, GUID_Friction:
                0x18    or    0x30
            Case GUID_CustomForce:
                0x10
        data
            Case GUID_ConstantForce:
                "lMagnitude" : "LONG"
            Case GUID_RampForce:
                "lStart" : "LONG"
                "lEnd" : "LONG"
            Case GUID_Square, GUID_Sine, GUID_Triangle, GUID_SawtoothUp, GUID_SawtoothDown:
                "dwMagnitude" : "DWORD"
                "lOffset" : "LONG"
                "dwPhase" : "DWORD"
                "dwPeriod" : "DWORD"
            Case GUID_Spring, GUID_Damper, GUID_Inertia, GUID_Friction:    (Warning: this effect can have double size)
                "lOffset" : "LONG"
                "lPositiveCoefficient" : "LONG"
                "lNegativeCoefficient" : "LONG"
                "dwPositiveSaturation" : "DWORD"
                "dwNegativeSaturation" : "DWORD"
                "lDeadBand" : "LONG"
            Case GUID_CustomForce:
                "cChannels" : "DWORD"
                "dwSamplePeriod" : "DWORD"
                "cSamples" : "DWORD"
                "rglForceData" : "LONG"

    CustomForceDataChunk    (Warning: this is hypothetical, not tested because FEdit can't generate CustomForce effects)
        label
            "data"
        data
            "rglForceData" : "LONG*"

"""



##########################     File Format Implementation and Verification (tests)    ##########################

class FFEFile(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "RIFF", noLengthInHeader = True, lastDataTypeIsArray = True,
                dataTypes = {
                                "header1"                       : "CHAR[0x4]",
                                "ffeHeader"                     : FFEHeader,
                                "fileEffects"                   : FileEffectChunk
                            },
                dataTypeOrder = ["header1","ffeHeader","fileEffects"] )
    def read(self, inp):
        self.expectedLength = len(inp) - 4
        offset = RIFFChunk.read(self, inp, 0)
        """ Tests """
        self.checkMember("header1", "\x00\x00\x00\x00") # Observation value
        return offset

class FFEHeader(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "FORC", noLengthInHeader = True, length = 0x10,
                dataTypes = {
                                "header2"                       : "CHAR[0x10]"
                            },
                dataTypeOrder = ["header2"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        """ Tests """
        self.checkMember("header2", "\x5C\x77\x7E\x19"+"\xBA\x34\xD3\x11"+"\xAB\xD5\x00\xC0"+"\x4F\x8E\xC6\x27") # Observation value
        return offset

class FileEffectChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "LIST",
                dataTypes = {
                                "effectStringChunk"             : EffectStringChunk,
                                "lpDiEffect"                    : EffectChunk,
                                "GuidEffect"                    : "CHAR[0x10]",
                                "unknownDWORD1"                 : "DWORD",
                                "axesChunk"                     : AxesChunk,
                                "directionChunk"                : DirectionChunk,
                                "envelopeChunk"                 : EnvelopeChunk,
                                "effectTypeSpecificChunk"       : EffectTypeSpecificChunk,
                              # "customForceDataChunk"          : CustomForceDataChunk # Warning: this last one is hypothetical, not tested because FEdit can't generate CustomForce effects
                            },
                dataTypeOrder = ["effectStringChunk","lpDiEffect","GuidEffect","unknownDWORD1","axesChunk","directionChunk","envelopeChunk","effectTypeSpecificChunk"] )#,"customForceDataChunk"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        if self.data["lpDiEffect"].data["lpvTypeSpecificParams"] != NULL:
            self.data["effectTypeSpecificChunk"].unpackData(self.data["GuidEffect"])
        """ Tests """
        self.checkMember("unknownDWORD1", 0x01) # Observation value
        if self.data["lpDiEffect"].data["lpvTypeSpecificParams"] != NULL:
            self.check( self.data["lpDiEffect"].data["cbTypeSpecificParams"] == self.data["effectTypeSpecificChunk"].length, 
                        '%s == data["cbTypeSpecificParams"] != data["effectTypeSpecificChunk"] == %s' % 
                                (self.data["lpDiEffect"].data["cbTypeSpecificParams"], self.data["effectTypeSpecificChunk"].length) )
        self.check( self.data["lpDiEffect"].data["cAxes"] * 0x04 == self.data["axesChunk"].length, 
                    '%s == data["cAxes"] * 0x04 != data["axesChunk"].length == %s' % 
                            (self.data["lpDiEffect"].data["cAxes"] * 0x04, self.data["axesChunk"].length) )
        self.check( self.data["lpDiEffect"].data["cAxes"] * 0x04 == self.data["directionChunk"].length, 
                    '%s == data["cAxes"] * 0x04 != data["directionChunk"].length == %s' % 
                            (self.data["lpDiEffect"].data["cAxes"] * 0x04, self.data["directionChunk"].length) )
        if self.data["lpDiEffect"].data["lpEnvelope"] != NULL:
            dwDuration = self.data["lpDiEffect"].data["dwDuration"]
            dwAttackTime = self.data["envelopeChunk"].data["dwAttackTime"]
            dwFadeTime = self.data["envelopeChunk"].data["dwFadeTime"]
            self.check( (dwDuration == -1) or (dwAttackTime + dwFadeTime <= dwDuration), 
                        '(%s == dwDuration != -1) and (%s == dwAttackTime + dwFadeTime > dwDuration)' % 
                                (dwDuration, dwAttackTime + dwFadeTime) )
        if self.data["lpDiEffect"].data["lpvTypeSpecificParams"] != NULL:
            if self.data["effectTypeSpecificChunk"].guid in ("GUID_Spring", "GUID_Damper", "GUID_Inertia", "GUID_Friction"):
                self.check( self.data["effectTypeSpecificChunk"].amountOfConditionStructs in (1, self.data["lpDiEffect"].data["cAxes"]), 
                            '%s == data["effectTypeSpecificChunk"].amountOfConditionStructs is not an element of (1, data["cAxes"] == %s)' % 
                                    (self.data["effectTypeSpecificChunk"].amountOfConditionStructs, self.data["lpDiEffect"].data["cAxes"]) )
            elif self.data["effectTypeSpecificChunk"].guid in ("GUID_Spring", "GUID_Damper", "GUID_Inertia", "GUID_Friction"):
                # if "customForceDataChunk" in self.data:
                self.check( self.data["effectTypeSpecificChunk"].data["cSamples"] == len(self.data["customForceDataChunk"].data["rglForceData"]), 
                            '%s == data["cSamples"] != len(data["rglForceData"]) == %s' % 
                                    (self.data["effectTypeSpecificChunk"].data["cSamples"], len(self.data["customForceDataChunk"].data["rglForceData"])) )
        return offset

class EffectStringChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "efct", noLengthInHeader = True, length = 0x40,
                dataTypes = {
                                "szFriendlyName"                : "CHAR[0x40]"
                            },
                dataTypeOrder = ["szFriendlyName"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        """ Tests """
        zeroTerminator = self.data["szFriendlyName"].find('\x00')
        self.check(zeroTerminator > -1, "data['szFriendlyName'] is not zero terminated")
        # Remove zero terminator char and possible memory garbage after the terminator char
        if zeroTerminator > -1: self.data["szFriendlyName"] = self.data["szFriendlyName"][:zeroTerminator]
        return offset

class EffectChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                noLengthInHeader = True, length = 0x38,
                dataTypes = {
                                "dwSize"                        : "DWORD",
                                "dwFlags"                       : "DWORD",
                                "dwDuration"                    : "DWORD",
                                "dwSamplePeriod"                : "DWORD",
                                "dwGain"                        : "DWORD",
                                "dwTriggerButton"               : "DWORD",
                                "dwTriggerRepeatInterval"       : "DWORD",
                                "cAxes"                         : "DWORD",
                                "rgdwAxes"                      : "LONG",
                                "rglDirection"                  : "LONG",
                                "lpEnvelope"                    : "LONG",
                                "cbTypeSpecificParams"          : "DWORD",
                                "lpvTypeSpecificParams"         : "LONG",
                                "dwStartDelay"                  : "DWORD" # if(DIRECTINPUT_VERSION >= 0x0600), this is satisfied because if(DIRECTINPUT_VERSION >= 0x0700) DIFILEEFFECT => FFE fileformat always contains "dwStartDelay"
                            },
                dataTypeOrder = ["dwSize","dwFlags","dwDuration","dwSamplePeriod","dwGain","dwTriggerButton","dwTriggerRepeatInterval","cAxes","rgdwAxes","rglDirection","lpEnvelope","cbTypeSpecificParams","lpvTypeSpecificParams","dwStartDelay"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        if self.data["lpEnvelope"] == NULL: # Then FileEffectChunk.data["envelopeChunk"] should not exist
            self.parent.excludedMembers.append("envelopeChunk")
        if self.data["lpvTypeSpecificParams"] == NULL: # Then FileEffectChunk.data["effectTypeSpecificChunk"] should not exist
            self.parent.excludedMembers.append("effectTypeSpecificChunk")
        """ Tests """
        self.checkMember("dwSize", self.length)
        # TODO: Validate dwFlags
        self.checkMemberRange("dwDuration", -1, 2**32 / 2 - 1) # strictly does not have a bound on range, but we check for positive and INFINITE (-1) numbers
        self.checkMemberRange("dwSamplePeriod", 0, 1e6) # strictly does not have a bound on range, but we check for positive numbers that are smaller than a second
        #self.checkMember("dwSamplePeriod", 0) # Curiosity: are there effects in my big FFE collection that explicitly use a custom sample period ?
        self.checkMemberRange("dwGain", 0, 10000)
        #self.checkMember("dwGain", 10000) # Curiosity: are there effects in my big FFE collection that use a gain!=max ?
        # TODO: Validate dwTriggerButton
        #self.checkMember("dwTriggerButton", -1) # Curiosity: are there effects in my big FFE collection that use a trigger button ?
        self.checkMemberRange("dwTriggerRepeatInterval", -1, 2**32 / 2 - 1) # strictly does not have a bound on range, but we check for positive and INFINITE (-1) numbers
        #self.checkMember("dwTriggerRepeatInterval", -1) # Curiosity: are there effects in my big FFE collection that use a triggerB interval != -1 ?
        self.checkMemberRange("cAxes", 1, 32)
        #self.checkMemberRange("cAxes", 0, 2) # Curiosity: are there effects in my big FFE collection that use more than 2 axes ?
        self.checkMember("rgdwAxes", 0x20) # Observation value
        self.checkMember("rglDirection", 0x40) # Observation value
        if self.data["lpEnvelope"] != NULL: self.checkMember("lpEnvelope", 0x80) # Observation value
        # Outside this chunk where EffectTypeSpecificChunk.length is available, check for equality with "cbTypeSpecificParams"
        if self.data["lpvTypeSpecificParams"] == NULL:
            self.checkMember("cbTypeSpecificParams", 0) # If this chuck doesn't exist, it should have zero size (not tested)
        else:
            self.checkMember("lpvTypeSpecificParams", 0x100) # Observation value
        self.checkMemberRange("dwStartDelay", 0, 2**32 / 2 - 1) # strictly does not have a bound on range, but we check for positive numbers
        return offset

class AxesChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "data",
                dataTypes = {
                                "rgdwAxes"                      : "DWORD*"
                            },
                dataTypeOrder = ["rgdwAxes"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        """ Tests """
        # TODO: Validate rgdwAxes
        # Outside this chunk where "cAxes" is available, check for self.length == cAxes * 0x4
        return offset

class DirectionChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "data",
                dataTypes = {
                                "rglDirection"                  : "LONG*"
                            },
                dataTypeOrder = ["rglDirection"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        """ Tests """
        # TODO: Validate rglDirection
        # Outside this chunk where "cAxes" is available, check for self.length == cAxes * 0x4
        return offset

class EnvelopeChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "data", length = 0x14,
                dataTypes = {
                                "dwSize"                        : "DWORD",
                                "dwAttackLevel"                 : "DWORD",
                                "dwAttackTime"                  : "DWORD",
                                "dwFadeLevel"                   : "DWORD",
                                "dwFadeTime"                    : "DWORD"
                            },
                dataTypeOrder = ["dwSize","dwAttackLevel","dwAttackTime","dwFadeLevel","dwFadeTime"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        """ Tests """
        self.checkMember("dwSize", self.length)
        self.checkMemberRange("dwAttackLevel", 0, 10000)
        self.checkMemberRange("dwFadeLevel", 0, 10000)
        # Outside this chunk where "dwDuration" is available, check for (("dwDuration" == -1) || ("dwAttackTime"+"dwFadeTime" <= "dwDuration"))
        return offset

class EffectTypeSpecificChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "data", dontUnpackData = True )
    def unpackData(self, guid):
        """ First set GUID specific length and datatypes """
        try:
            self.guid = GUIDs[guid]
        except KeyError:
            self.guid = "UNKNOWN_GUID"
        self.setAdditionalRepresentation(self.guid)
        
        if self.guid == "GUID_ConstantForce":
            self.expectedLength = 0x4
            dataTypes = {
                                "lMagnitude"                    : "LONG"
                        }
            dataTypeOrder = ["lMagnitude"]
        
        elif self.guid == "GUID_RampForce":
            self.expectedLength = 0x8
            dataTypes = {
                                "lStart"                        : "LONG",
                                "lEnd"                          : "LONG"
                        }
            dataTypeOrder = ["lStart","lEnd"]
        
        elif self.guid in ("GUID_Square", "GUID_Sine", "GUID_Triangle", "GUID_SawtoothUp", "GUID_SawtoothDown"):
            self.expectedLength = 0x10
            dataTypes = {
                                "dwMagnitude"                   : "DWORD",
                                "lOffset"                       : "LONG",
                                "dwPhase"                       : "DWORD",
                                "dwPeriod"                      : "DWORD"
                        }
            dataTypeOrder = ["dwMagnitude","lOffset","dwPhase","dwPeriod"]
        
        elif self.guid in ("GUID_Spring", "GUID_Damper", "GUID_Inertia", "GUID_Friction"):
            self.amountOfConditionStructs = int(round(self.length / float(0x18))) # Multiple condition structs are possible: one for each axis.
            self.expectedLength = self.amountOfConditionStructs * 0x18
            # Outside this chunk where "cAxes" is available, check for (amountOfConditionStructs in (1, cAxes))
            dataTypes = {
                                "lOffset"                       : "LONG",
                                "lPositiveCoefficient"          : "LONG",
                                "lNegativeCoefficient"          : "LONG",
                                "dwPositiveSaturation"          : "DWORD",
                                "dwNegativeSaturation"          : "DWORD",
                                "lDeadBand"                     : "LONG"
                        }
            dataTypeOrder = ["lOffset","lPositiveCoefficient","lNegativeCoefficient","dwPositiveSaturation","dwNegativeSaturation","lDeadBand"]
            if self.amountOfConditionStructs > 1: # If more than 1 struct: give them an index.
                dataTypes_new     = {"%s[%s]" % (member, i) : dataTypes[member] for i in range(self.amountOfConditionStructs) for member in dataTypeOrder}
                dataTypeOrder_new = ["%s[%s]" % (member, i)                     for i in range(self.amountOfConditionStructs) for member in dataTypeOrder]
                dataTypes, dataTypeOrder = dataTypes_new, dataTypeOrder_new
        
        elif self.guid == "GUID_CustomForce":
            self.expectedLength = 0x10
            dataTypes = {
                                "cChannels"                     : "DWORD",
                                "dwSamplePeriod"                : "DWORD",
                                "cSamples"                      : "DWORD",
                                "rglForceData"                  : "LONG"
                        }
            dataTypeOrder = ["cChannels","dwSamplePeriod","cSamples","rglForceData"]
        
        else:
            self.setDataTypes({}, [])
            return self.alert("Unknown GUID: %s" % repr(guid))
        """ Unpack data """
        self.check( self.length == self.expectedLength, 
                    "at inp[%s:%s+4]: Chunk length == %s != %s" % 
                            (self.globalOffset-4, self.globalOffset-4, self.length, self.expectedLength) )
        self.setDataTypes(dataTypes, dataTypeOrder)
        self.dontUnpackData = False
        RIFFChunk.unpackData(self)
        
        """ Tests """
        if self.guid == "GUID_ConstantForce":
            self.checkMemberRange("lMagnitude", -10000, 10000)
        
        elif self.guid == "GUID_RampForce":
            self.checkMemberRange("lStart", -10000, 10000)
            self.checkMemberRange("lEnd", -10000, 10000)
        
        elif self.guid in ("GUID_Square", "GUID_Sine", "GUID_Triangle", "GUID_SawtoothUp", "GUID_SawtoothDown"):
            self.checkMemberRange("dwMagnitude", 0, 10000)
            self.checkMemberRange("lOffset", -10000, 10000)
            self.checkMemberRange("dwPhase", 0, 35999)
            self.checkMemberRange("dwPeriod", 0, 2**32 / 2 - 1) # does not have a bound on range, but we check for strictly positive numbers
            self.checkMemberNot("dwPeriod", 0) # strictly positive number

        elif self.guid in ("GUID_Spring", "GUID_Damper", "GUID_Inertia", "GUID_Friction"):
            for i in range(self.amountOfConditionStructs):
                self.checkMemberRange(self.dataTypeOrder[i*6 + 0], -10000, 10000) # "lOffset"
                self.checkMemberRange(self.dataTypeOrder[i*6 + 1], -10000, 10000) # "lPositiveCoefficient"
                self.checkMemberRange(self.dataTypeOrder[i*6 + 2], -10000, 10000) # "lNegativeCoefficient"
                self.checkMemberRange(self.dataTypeOrder[i*6 + 3],      0, 10000) # "dwPositiveSaturation"
                self.checkMemberRange(self.dataTypeOrder[i*6 + 4],      0, 10000) # "dwNegativeSaturation"
                self.checkMemberRange(self.dataTypeOrder[i*6 + 5],      0, 10000) # "lDeadBand"
        
        elif self.guid == "GUID_CustomForce":
            self.checkMemberRange("cChannels", 1, 32)
            self.checkMemberRange("dwSamplePeriod", 0, 2**32 / 2 - 1) # does not have a bound on range, but we check for strictly positive numbers
            self.checkMemberNot("dwSamplePeriod", 0) # strictly positive number
            self.checkMemberRange("cSamples", 0, 2**32 / 2 - 1)
            self.check( not (self.data["cSamples"] % self.data["cChannels"]), # cSamples integral multiple of cChannels
                        '(%s = data["cSamples"]) %% (data["cChannels"] = %s) == %s != 0' %
                                (self.data["cSamples"], data["cChannels"], self.data["cSamples"] % self.data["cChannels"]) )
            self.checkMember("rglForceData", 0x200) # Hypothesis, not encountered yet

class CustomForceDataChunk(RIFFChunk):
    def __init__(self):
        RIFFChunk.__init__( self,
                label = "data",
                dataTypes = {
                                "rglForceData"                  : "LONG*"
                            },
                dataTypeOrder = ["rglForceData"] )
    def read(self, inp, offset, globalOffset = 0):
        offset = RIFFChunk.read(self, inp, offset, globalOffset)
        """ Tests """
        # TODO: Validate rglForceData
        for i, f in enumerate(self.data["rglForceData"]):
            self.check( -10000 <= f <= 10000,
                        'data["rglForceData"][%s] == %s, so is not an element in range(-10000, 10000)' %
                                (i, f) )
        # Outside this chunk where "cSamples" is available, check for len(self.data["rglForceData"]) == "cSamples"
        return offset





##########################     Main    ##########################

def main():
    import os, sys
    
    if len(sys.argv) < 2:
        print ("This program will try to open M$ FFE (force feedback) files, and validate/verify the file.")
        print ("\t They could have been made with the FEdit tool (from an old DX SDK).\n")
        print ("Usage: \t python FFE_RevEng.py <filename>")
        print ("or:\t python FFE_RevEng.py <directory containing FFE files>")

    else:
        def readFFE(fileName):
            f = FFEFile()
            #inp = open("/home/elias/VirtualBoxSharedFolders/FFE_ReverseEngineer/Empty.ffe", 'rb').read()
            inp = open(fileName, 'rb').read()
            print ( "\n\nFile: '%s'\nLength of file: %s bytes.\n%s" % (fileName, len(inp), '-'*60) )
            f.read(inp)
            print ("\nDumping File: ")
            print (f.reprUnpackedData())
        
        fileName = sys.argv[1]
        if os.path.isdir(fileName):
            for fN in sorted(os.listdir(fileName)):
                ffeFN = os.path.join(fileName, fN)
                if ffeFN.endswith('.ffe') and os.path.isfile(ffeFN):
                    try:
                        readFFE(ffeFN)
                    except:
                        pass # Pass to give a chance for other FFE files to be parsed
        else:
            readFFE(fileName)


if __name__ == "__main__":
    main()
